import { compose, createStore, applyMiddleware } from 'redux';
import memoizerific from 'memoizerific';
import produce, { setAutoFreeze } from 'immer';
import reduxThunk from 'redux-thunk';
import React, { createContext, useContext, useState, useRef, useEffect } from 'react';
import shallowEqual from 'shallowequal';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

var isStateObject = function isStateObject(x) {
  return x !== null && typeof x === 'object' && !Array.isArray(x);
};

/**
 * immer is an implementation detail, so we are not going to use its auto freeze
 * behaviour, which throws errors if trying to mutate state. It's also risky
 * for production builds as has a perf overhead.
 *
 * @see https://github.com/mweststrate/immer#auto-freezing
 */

setAutoFreeze(false);
var maxSelectFnMemoize = 100;
var actionSymbol = '__action__';
var actionNameSymbol = '__actionName__';
var thunkSymbol = '__thunk__';
var listenSymbol = '__listen__';
var metaSymbol = '__meta__';
var selectSymbol = '__select__';
var selectImpSymbol = '__selectImp__';
var selectDependenciesSymbol = '__selectDependencies__';
var selectStateSymbol = '__selectState__';
var reducerSymbol = '__reducer__';

var get = function get(path, target) {
  return path.reduce(function (acc, cur) {
    return isStateObject(acc) ? acc[cur] : undefined;
  }, target);
};

var set$1 = function set(path, target, value) {
  path.reduce(function (acc, cur, idx) {
    if (idx + 1 === path.length) {
      acc[cur] = value;
    } else {
      acc[cur] = acc[cur] || {};
    }

    return acc[cur];
  }, target);
};

var tick = function tick() {
  return new Promise(function (resolve) {
    return setTimeout(resolve);
  });
};

var helpers = {
  actionName: function actionName(action) {
    return action[actionNameSymbol];
  },
  thunkStartName: function thunkStartName(action) {
    return action[actionNameSymbol] + "(started)";
  },
  thunkCompleteName: function thunkCompleteName(action) {
    return action[actionNameSymbol] + "(completed)";
  },
  action: function action(fn) {
    fn[actionSymbol] = true;
    return fn;
  },
  thunk: function thunk(fn) {
    fn[thunkSymbol] = true;
    return fn;
  },
  select: function select(fn, dependencies) {
    fn[selectSymbol] = true;
    fn[selectDependenciesSymbol] = dependencies;
    fn[selectStateSymbol] = {};
    return fn;
  },
  reducer: function reducer(fn) {
    fn[reducerSymbol] = true;
    return fn;
  },
  listen: function listen(fn) {
    fn[listenSymbol] = true;
    return fn;
  }
};
var createStore$1 = function createStore$$1(model, options) {
  if (options === void 0) {
    options = {};
  }

  var _options = options,
      compose$$1 = _options.compose,
      _options$devTools = _options.devTools,
      devTools = _options$devTools === void 0 ? true : _options$devTools,
      _options$disableInter = _options.disableInternalSelectFnMemoize,
      disableInternalSelectFnMemoize = _options$disableInter === void 0 ? false : _options$disableInter,
      _options$initialState = _options.initialState,
      initialState = _options$initialState === void 0 ? {} : _options$initialState,
      injections = _options.injections,
      _options$mockActions = _options.mockActions,
      mockActions = _options$mockActions === void 0 ? false : _options$mockActions,
      _options$middleware = _options.middleware,
      middleware = _options$middleware === void 0 ? [] : _options$middleware,
      _options$reducerEnhan = _options.reducerEnhancer,
      reducerEnhancer = _options$reducerEnhan === void 0 ? function (rootReducer) {
    return rootReducer;
  } : _options$reducerEnhan;

  var wrapFnWithMemoize = function wrapFnWithMemoize(x) {
    return !disableInternalSelectFnMemoize && typeof x === 'function' ? memoizerific(maxSelectFnMemoize)(x) : x;
  };

  var definition = _extends({}, model, {
    logFullState: helpers.thunk(function (actions, payload, _ref) {
      var getState = _ref.getState;
      // eslint-disable-next-line no-console
      console.log(JSON.stringify(getState(), null, 2));
    })
  });

  var references = {};
  var defaultState = {};
  var actionThunks = {};
  var actionCreators = {};
  var actionCreatorDict = {};
  var customReducers = [];
  var selectorReducers = [];
  var listenDefinitions = [];
  var thunkListenersDict = {};
  var actionListenersDict = {};
  var actionReducersDict = {};

  var dispatchThunkListeners = function dispatchThunkListeners(name, payload) {
    var listensForAction = thunkListenersDict[name];
    return listensForAction && listensForAction.length > 0 ? Promise.all(listensForAction.map(function (listenForAction) {
      return listenForAction(get(listenForAction[metaSymbol].parent, actionCreators), payload, {
        dispatch: references.dispatch,
        getState: references.getState,
        injections: injections,
        meta: listenForAction[metaSymbol]
      });
    })) : Promise.resolve();
  };

  var extract = function extract(current, parentPath) {
    return Object.keys(current).forEach(function (key) {
      var value = current[key];
      var path = [].concat(parentPath, [key]);
      var meta = {
        parent: parentPath,
        path: path
      };

      if (typeof value === 'function') {
        if (value[actionSymbol]) {
          var name = "@action." + path.join('.');
          value[actionNameSymbol] = name;
          value[metaSymbol] = meta; // Action Reducer

          var actionReducer = value;
          actionReducer[actionNameSymbol] = name;
          actionReducersDict[name] = actionReducer;

          var actionCreator = function actionCreator(payload) {
            var result = references.dispatch({
              type: actionReducer[actionNameSymbol],
              payload: payload
            });
            return result;
          };

          actionCreator[actionNameSymbol] = name;
          actionCreatorDict[name] = actionCreator;
          set$1(path, actionCreators, actionCreator);
        } else if (value[selectSymbol]) {
          // skip
          value[selectStateSymbol] = {
            parentPath: parentPath,
            key: key,
            executed: false
          };
          selectorReducers.push(value);
        } else if (value[thunkSymbol]) {
          var _name = "@thunk." + path.join('.');

          value[actionNameSymbol] = _name; // Thunk Action

          var action = function action(payload) {
            return value(get(parentPath, actionCreators), payload, {
              dispatch: references.dispatch,
              getState: references.getState,
              injections: injections,
              meta: meta
            });
          };

          set$1(path, actionThunks, action); // Thunk Action Creator

          var _actionCreator = function _actionCreator(payload) {
            return tick().then(function () {
              return references.dispatch({
                type: _name + "(started)",
                payload: payload
              });
            }).then(function () {
              return references.dispatch(function () {
                return action(payload);
              });
            }).then(function (result) {
              references.dispatch({
                type: _name + "(completed)",
                payload: payload
              });
              return result;
            }).catch(function (err) {
              references.dispatch({
                type: _name + "(failed)",
                payload: err
              });
              return Promise.reject(err);
            });
          };

          _actionCreator[actionNameSymbol] = _name;
          actionCreatorDict[_name] = _actionCreator;
          set$1(path, actionCreators, _actionCreator);
        } else if (value[reducerSymbol]) {
          customReducers.push({
            path: path,
            reducer: value
          });
        } else if (value[listenSymbol]) {
          listenDefinitions.push(value);
          value[metaSymbol] = meta;
        } else {
          // eslint-disable-next-line no-console
          console.warn("Easy Peasy: Found a function at path " + path.join('.') + " in your model. Version 2 required that you wrap functions with the action helper");
        }
      } else if (isStateObject(value) && Object.keys(value).length > 0) {
        extract(value, path);
      } else {
        // State
        var initialParentRef = get(parentPath, initialState);

        if (initialParentRef && key in initialParentRef) {
          set$1(path, defaultState, initialParentRef[key]);
        } else {
          set$1(path, defaultState, value);
        }
      }
    });
  };

  extract(definition, []);
  selectorReducers.forEach(function (selector) {
    selector[selectImpSymbol] = function (state) {
      return wrapFnWithMemoize(selector(state));
    };
  });
  listenDefinitions.forEach(function (def) {
    var on = function on(target, handler) {
      if (typeof handler !== 'function') {
        return;
      }

      var meta = def[metaSymbol];
      handler[metaSymbol] = meta;
      var name;

      if (typeof target === 'function' && target[actionNameSymbol] && actionCreatorDict[target[actionNameSymbol]]) {
        if (target[thunkSymbol]) {
          name = helpers.thunkCompleteName(target);
        } else {
          name = target[actionNameSymbol];
        }
      } else if (typeof target === 'string') {
        name = target;
      }

      if (name) {
        if (handler[thunkSymbol]) {
          thunkListenersDict[name] = thunkListenersDict[name] || [];
          thunkListenersDict[name].push(handler);
        } else {
          actionListenersDict[name] = actionListenersDict[name] || [];
          actionListenersDict[name].push({
            path: meta.parent,
            handler: handler
          });
        }
      }
    };

    def(on);
  });

  var createReducers = function createReducers() {
    var runActionReducerAtPath = function runActionReducerAtPath(state, action, actionReducer, path) {
      var current = get(path, state);

      if (path.length === 0) {
        return produce(state, function (_draft) {
          return actionReducer(_draft, action.payload);
        });
      }

      return produce(state, function (draft) {
        set$1(actionReducer[metaSymbol].parent, draft, produce(current, function (_draft) {
          return actionReducer(_draft, action.payload);
        }));
      });
    };

    var reducerForActions = function reducerForActions(state, action) {
      var actionReducer = actionReducersDict[action.type];

      if (actionReducer) {
        return runActionReducerAtPath(state, action, actionReducer, actionReducer[metaSymbol].parent);
      }

      return state;
    };

    var reducerForListeners = function reducerForListeners(state, action) {
      var actionListeners = actionListenersDict[action.type];

      if (actionListeners) {
        return actionListeners.reduce(function (newState, _ref2) {
          var path = _ref2.path,
              handler = _ref2.handler;
          return runActionReducerAtPath(newState, action, handler, path);
        }, state);
      }

      return state;
    };

    var reducerForCustomReducers = function reducerForCustomReducers(state, action) {
      return produce(state, function (draft) {
        customReducers.forEach(function (_ref3) {
          var p = _ref3.path,
              red = _ref3.reducer;
          var current = get(p, draft);
          set$1(p, draft, red(current, action));
        });
      });
    };

    var runSelectorReducer = function runSelectorReducer(state, selector) {
      var _selector$selectState = selector[selectStateSymbol],
          parentPath = _selector$selectState.parentPath,
          key = _selector$selectState.key,
          executed = _selector$selectState.executed;

      if (executed) {
        return state;
      }

      var dependencies = selector[selectDependenciesSymbol];
      var stateAfterDependencies = dependencies ? dependencies.reduce(runSelectorReducer, state) : state;
      var newState = stateAfterDependencies;

      if (parentPath.length > 0) {
        var target = get(parentPath, stateAfterDependencies);

        if (target) {
          if (!selector.prevState || selector.prevState !== state) {
            var newValue = selector[selectImpSymbol](target);
            newState = produce(state, function (draft) {
              var updateTarget = get(parentPath, draft);
              updateTarget[key] = newValue;
            });
            selector.prevState = newState;
          }
        }
      } else if (!selector.prevState || selector.prevState !== state) {
        var _newValue = selector[selectImpSymbol](stateAfterDependencies);

        newState = produce(state, function (draft) {
          draft[key] = _newValue;
        });
        selector.prevState = newState;
      }

      selector[selectStateSymbol].executed = true;
      return newState;
    };

    var runSelectors = function runSelectors(state) {
      return selectorReducers.reduce(runSelectorReducer, state);
    };

    var isInitial = true;

    var selectorsReducer = function selectorsReducer(state) {
      var stateAfterSelectors = runSelectors(state);
      isInitial = false;
      selectorReducers.forEach(function (selector) {
        selector[selectStateSymbol].executed = false;
      });
      return stateAfterSelectors;
    };

    return function (state, action) {
      var stateAfterActions = reducerForActions(state, action);
      var stateAfterListeners = reducerForListeners(stateAfterActions, action);
      var stateAfterCustomReducers = reducerForCustomReducers(stateAfterListeners, action);

      if (state !== stateAfterCustomReducers || isInitial) {
        return selectorsReducer(stateAfterCustomReducers);
      }

      return stateAfterCustomReducers;
    };
  };

  var reducers = createReducers();
  var composeEnhancers = compose$$1 || (devTools && typeof window !== 'undefined' && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ : compose);

  var dispatchActionStringListeners = function dispatchActionStringListeners() {
    return function (next) {
      return function (action) {
        if (thunkListenersDict[action.type]) {
          dispatchThunkListeners(action.type, action.payload);
        }

        return next(action);
      };
    };
  };

  var mockedActions = [];

  var mockActionsMiddlware = function mockActionsMiddlware() {
    return function (next) {
      return function (action) {
        if (mockActions) {
          mockedActions.push(action);
          return undefined;
        }

        return next(action);
      };
    };
  };

  var store = createStore(reducerEnhancer(reducers), defaultState, composeEnhancers(applyMiddleware.apply(void 0, [reduxThunk, dispatchActionStringListeners, mockActionsMiddlware].concat(middleware))));

  store.getMockedActions = function () {
    return [].concat(mockedActions);
  };

  store.clearMockedActions = function () {
    mockedActions = [];
  }; // attach the action creators to dispatch


  Object.keys(actionCreators).forEach(function (key) {
    store.dispatch[key] = actionCreators[key];
  });
  references.dispatch = store.dispatch;
  references.getState = store.getState;
  references.getState.getState = store.getState;
  return store;
};

var StoreContext = createContext();

function useStore(mapState, dependencies) {
  if (dependencies === void 0) {
    dependencies = [];
  }

  var store = useContext(StoreContext);

  var _useState = useState(mapState(store.getState())),
      state = _useState[0],
      setState = _useState[1]; // As our effect only fires on mount and unmount it won't have the state
  // changes visible to it, therefore we use a mutable ref to track this.


  var stateRef = useRef(state); // Helps avoid firing of events when unsubscribed, i.e. unmounted

  var isActive = useRef(true);
  useEffect(function () {
    var calculateState = function calculateState() {
      var newState = mapState(store.getState());
      isActive.current = true;

      if (newState === stateRef.current || isStateObject(newState) && isStateObject(stateRef.current) && shallowEqual(newState, stateRef.current)) {
        // Do nothing
        return;
      }

      stateRef.current = newState;

      if (isActive.current) {
        setState(stateRef.current);
      }
    };

    calculateState();
    var unsubscribe = store.subscribe(calculateState);
    return function () {
      unsubscribe();
      isActive.current = false;
    };
  }, dependencies);
  return state;
}
function useActions(mapActions) {
  var store = useContext(StoreContext);
  return mapActions(store.dispatch);
}
function useDispatch() {
  var store = useContext(StoreContext);
  return store.dispatch;
}
function createTypedHooks() {
  return {
    useActions: useActions,
    useDispatch: useDispatch,
    useStore: useStore
  };
}

var StoreProvider = function StoreProvider(_ref) {
  var children = _ref.children,
      store = _ref.store;
  return React.createElement(StoreContext.Provider, {
    value: store
  }, children);
};

var action = helpers.action,
    actionName = helpers.actionName,
    listen = helpers.listen,
    reducer = helpers.reducer,
    select = helpers.select,
    thunk = helpers.thunk,
    thunkStartName = helpers.thunkStartName,
    thunkCompleteName = helpers.thunkCompleteName;

export { action, actionName, createStore$1 as createStore, createTypedHooks, listen, reducer, select, StoreProvider, thunk, thunkStartName, thunkCompleteName, useActions, useDispatch, useStore };
//# sourceMappingURL=easy-peasy.esm.js.map
